
#include "libsangoma.h"
#include <sys/mman.h>
#include <pthread.h>
//#include "lib_api.h"		

#define MAX_CHANS 255

typedef struct sangoma_chan {
	int spanno;
	int channo;
	sng_fd_t dev_fd;
} sangoma_chan_t;
sangoma_chan_t sangoma_channels[MAX_CHANS];

static u_int32_t	poll_events_bitmap = 0;
static sangoma_wait_obj_t *sangoma_wait_objects[MAX_CHANS];

u_int32_t input_flags[MAX_CHANS];
u_int32_t output_flags[MAX_CHANS];
static u_int64_t timestampPOLLIN[MAX_CHANS] = {};
static u_int64_t timestampPOLLOUT[MAX_CHANS] = {};
int application_termination_flag = 0;

pthread_t timing_thread;
pthread_mutex_t g_lock;


/* This example application has only a single execution thread - it is safe
 * to use a global buffer for received data and for data to be transmitted. */
static unsigned char rxdata[MAX_NO_DATA_BYTES_IN_FRAME];
/*
// it's doing rx2tx, so rx buffer's data will be written to device directly

static unsigned char txdata[MAX_NO_DATA_BYTES_IN_FRAME];
*/

static int launch_timing_thread(void);
void *timing_func(void *obj);

void TerminateHandler (int sig);
void handle_span_chan(int open_device_counter);
static int log_printf(char *fmt, ...);


static int log_printf(char *fmt, ...)
{
    char *data;
    int ret = 0;
    va_list ap;
    char date[200] = "";
    struct tm now;
    time_t epoch;


#if 0
    if (verbose < 0) {
        /* Continue through */
    } else if (!verbose) {
            return 0;
    }
#endif
   
    if (time(&epoch) && localtime_r(&epoch, &now)) {
        strftime(date, sizeof(date), "%Y-%m-%d %T", &now);
    }

    va_start(ap, fmt);
    ret = vasprintf(&data, fmt, ap);
    if (ret == -1) {
        return ret;
    }

    pthread_mutex_lock(&g_lock);
    fprintf(stderr, "[%d] %s %s", getpid(), date, data);
    pthread_mutex_unlock(&g_lock);

    if (data) {
            free(data);
    }

    return 0;
}

void TerminateHandler (int sig)
{
	sng_fd_t fd;
	log_printf("\nProcess terminated by user request.\n");
	application_termination_flag = 1;

	fd = sangoma_wait_obj_get_fd(sangoma_wait_objects[0]);
	sangoma_close(&fd);
	return;
}




void handle_span_chan(int open_device_counter)
{
	int	iResult, i;
	int	Rx_lgth = 0;
	int	err = 0;
	u_int64_t timestamp, diff;
	struct timespec tp;
	wanpipe_api_t tdm_api;
	memset(&tdm_api,0,sizeof(tdm_api));

	wp_api_hdr_t	rxhdr;

	for (i=0;i<open_device_counter;i++) {
		sng_fd_t dev_fd = sangoma_wait_obj_get_fd(sangoma_wait_objects[i]);
		sangoma_flush_stats(dev_fd,&tdm_api);
		err = sangoma_writemsg(
					dev_fd,
					&rxhdr,			/* header buffer */
					sizeof(wp_api_hdr_t),	/* header size */
					rxdata,			/* data buffer */
					248,	/* DATA size */
					0);
		err = sangoma_writemsg(
					dev_fd,
					&rxhdr,			/* header buffer */
					sizeof(wp_api_hdr_t),	/* header size */
					rxdata,			/* data buffer */
					248,	/* DATA size */
					0);
		input_flags[i] = poll_events_bitmap;
	}

	/* Main Rx/Tx/Event loop */
	while(!application_termination_flag) 
	{	
		iResult = sangoma_waitfor_many(sangoma_wait_objects, 
					       input_flags,
					       output_flags,
				               open_device_counter /* number of wait objects */,
					       5000 /* wait timeout, in milliseconds */);
		switch(iResult)
		{
		case SANG_STATUS_APIPOLL_TIMEOUT:
			/* timeout (not an error) */
			log_printf("Timeout\n");
			continue;

		case SANG_STATUS_SUCCESS:
		    	clock_gettime(CLOCK_MONOTONIC, &tp);
            		timestamp = tp.tv_sec * 1000000000UL + tp.tv_nsec;
			
			for(i = 0; i < open_device_counter; i++){
				sng_fd_t dev_fd = sangoma_wait_obj_get_fd(sangoma_wait_objects[i]);
				sangoma_chan_t *chan = sangoma_wait_obj_get_context(sangoma_wait_objects[i]);

				if(output_flags[i] & POLLIN){
					
					diff = timestamp - timestampPOLLIN[i];
					if ((diff > 1500000UL || diff < 500000UL) && timestampPOLLIN[i]) {
						log_printf("Span: %d, Chan: %d: POLLIN delta= %lu.%06lu ms\n", chan->spanno, chan->channo, diff / 1000000, diff % 1000000);
					}

					/* rx data OR a free tx buffer available */
					Rx_lgth = sangoma_readmsg(
								dev_fd,
								&rxhdr,				/* header buffer */
								sizeof(wp_api_hdr_t),		/* header size */
								rxdata,				/* data buffer */
								MAX_NO_DATA_BYTES_IN_FRAME,	/* data BUFFER size */
								0);   

					if(Rx_lgth <= 0) {
						log_printf("Span: %d, Chan: %d: Error receiving data!\n", 
							chan->spanno, chan->channo);
					}
					timestampPOLLIN[i] = timestamp;
				}
				if(output_flags[i] & POLLOUT){
					diff = timestamp - timestampPOLLOUT[i];
					if ((diff > 1500000UL || diff < 500000UL) && timestampPOLLOUT[i]) {
						log_printf("Span: %d, Chan: %d: POLLOUT delta= %lu.%06lu ms\n", chan->spanno, chan->channo, diff / 1000000, diff % 1000000);
					}
					err = sangoma_writemsg(
								dev_fd,
								&rxhdr,			/* header buffer */
								sizeof(wp_api_hdr_t),	/* header size */
								rxdata,			/* data buffer */
								rxhdr.data_length,	/* DATA size */
								0);
					if (err <= 0){
						log_printf("Span: %d, Chan: %d: Failed to send %s!\n", 
							chan->spanno, 
							chan->channo, strerror(errno));
					}
					timestampPOLLOUT[i] = timestamp;
				}
			}/* for() */
			break;

		default:
			/* error */
			log_printf("Error: iResult: %s (%d)\n", SDLA_DECODE_SANG_STATUS(iResult), iResult);
			return;
		}

	}/* while() */
}


/*!
  \fn int __cdecl main(int argc, char* argv[])
  \brief Main function that starts the sample code
  \param argc  number of arguments
  \param argv  argument list
*/
int __cdecl main(int argc, char* argv[])
{
	int err;
	sng_fd_t	dev_fd;
	int span = 0;
	int chan = 1;
	int open_dev_cnt = 0;
	wanpipe_api_t tdm_api;
	sangoma_status_t status;
	struct sched_param sp;
	cpu_set_t my_set;        /* Define your cpu_set bit mask. */

	sangoma_wait_objects[0] = NULL;

	memset(&tdm_api,0,sizeof(tdm_api));

	pthread_mutex_init(&g_lock, NULL);


	poll_events_bitmap = 0;
	poll_events_bitmap |= POLLIN;
	poll_events_bitmap |= POLLOUT;
	/*
	// we do rx2tx, so not setting POLLOUT
	poll_events_bitmap |= POLLOUT;
	// we don't handle event in this sample, so not setting POLLPRI	
	poll_events_bitmap |= (POLLHUP | POLLPRI);
	*/

    	sp.sched_priority = sched_get_priority_max(SCHED_RR);
      	sched_setscheduler(0, SCHED_RR, &sp);
	mlockall(MCL_FUTURE);

	CPU_ZERO(&my_set);       /* Initialize it all to 0, i.e. no CPUs selected. */
	CPU_SET(3, &my_set);     /* set the bit that represents core 7. */
	//sched_setaffinity(0, sizeof(cpu_set_t), &my_set); /* Set affinity of tihs process to */
                                                  /* the defined mask, i.e. only 7. */

	signal(SIGHUP,TerminateHandler);
	signal(SIGTERM,TerminateHandler);
		
	log_printf("Sample 1.1\n");

	for (span=1;span<=8;span++) {

		log_printf("Connecting to Port/Span: %d, Interface/Chan: %d\n", span, chan);
		dev_fd = sangoma_open_api_span_chan(span, chan);
		sangoma_channels[open_dev_cnt].dev_fd = dev_fd;
		if( dev_fd == INVALID_HANDLE_VALUE){
			log_printf("Warning: Failed to open span %d, chan %d\n", span , chan);
			return 1;
		}else{
			log_printf("Successfuly opened span %d, chan %d\n", span , chan);
		}

		status = sangoma_wait_obj_create(&sangoma_wait_objects[open_dev_cnt], dev_fd, SANGOMA_DEVICE_WAIT_OBJ);
		if(status != SANG_STATUS_SUCCESS){
			log_printf("Error: Failed to create 'sangoma_wait_object'!\n");
			return 1;
		}

		sangoma_channels[open_dev_cnt].channo = chan;
		sangoma_channels[open_dev_cnt].spanno = span;
		sangoma_wait_obj_set_context(sangoma_wait_objects[open_dev_cnt], &sangoma_channels[open_dev_cnt]);
		
		if((err = sangoma_get_full_cfg(dev_fd, &tdm_api))){
			log_printf("Error: Failed to create 'tdm_api'!\n");
			return 1;
		}
		
		sangoma_set_tx_queue_sz(dev_fd,&tdm_api,2);
		sangoma_flush_bufs(dev_fd,&tdm_api);
		sangoma_flush_tx_bufs(dev_fd,&tdm_api);
		sangoma_flush_rx_bufs(dev_fd,&tdm_api);
		sangoma_flush_stats(dev_fd,&tdm_api);

		open_dev_cnt++;
	}

	launch_timing_thread();
	
	handle_span_chan(open_dev_cnt /* handle a single device */);

	return 0;
}

void *timing_func(void *obj)
{
	u_int64_t timestamp, timestamp_next, diff;
	struct timespec tp;

	while (1) {
		clock_gettime(CLOCK_MONOTONIC, &tp);
       	timestamp = tp.tv_sec * 1000000000UL + tp.tv_nsec;
		usleep(1000);
		clock_gettime(CLOCK_MONOTONIC, &tp);
       	timestamp_next = tp.tv_sec * 1000000000UL + tp.tv_nsec;
		diff = timestamp_next - timestamp;
		if (diff > 1500000UL || diff < 500000UL) {
			log_printf("Usleep delay delta= %lu.%06lu ms\n", diff / 1000000, diff % 1000000);
		}
	}	

	return NULL;
}

#define MGD_STACK_SIZE 1024*1024

static int launch_timing_thread()
{
        pthread_attr_t attr;
        int result = -1;

        result = pthread_attr_init(&attr);
        pthread_attr_setinheritsched(&attr, PTHREAD_EXPLICIT_SCHED);
        pthread_attr_setschedpolicy(&attr, SCHED_RR);
        pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
        pthread_attr_setstacksize(&attr, MGD_STACK_SIZE);

        result = pthread_create(&timing_thread, &attr, timing_func, (void*)NULL);
        if (result) {
                log_printf("%s: Error: Creating Thread! %s\n",
                                 __FUNCTION__,strerror(errno));
        }
        pthread_attr_destroy(&attr);

        return result;
}

